==> haystack-0.5.0%2B2.tar.gz: lib/src/io/HZincReader.dart <==
///
/// Copyright (c) 2018, BAS Services & Graphics, LLC.
/// Licensed under the Academic Free License version 3.0
///

library haystack.hzincreader;

import 'package:unicode_helper/unicode_helper.dart' as unicode;

import '../HBin.dart';
import '../HBool.dart';
import '../HCoord.dart';
import '../HDictBuilder.dart';
import '../HGrid.dart';
import '../HGridBuilder.dart';
import '../HList.dart';
import '../HMarker.dart';
import '../HNA.dart';
import '../HNum.dart';
import '../HRef.dart';
import '../HRemove.dart';
import '../HStr.dart';
import '../HVal.dart';
import '../HXStr.dart';
import '../ParseError.dart';
import 'HGridReader.dart';
import 'HaystackToken.dart';
import 'HaystackTokenizer.dart';

/// HZincReader reads grids using the Zinc format.
///
/// See [Project Haystack](http://project-haystack.org/doc/Zinc)
class HZincReader implements HGridReader {

  HZincReader(String s) :
        this._tokenizer = new HaystackTokenizer(s.runes) {
    _init();
  }

  void _init() {
    _consume();
    _consume();
  }

  /// Read a value.
  HVal readVal() {
    HVal val = null;
    if (_cur == HaystackToken.id && "ver" == _curVal)
      val = _parseGrid();
    else
      val = _parseVal();
    _verify(HaystackToken.eof);
    return val;
  }

//////////////////////////////////////////////////////////////////////////
// HGridReader
//////////////////////////////////////////////////////////////////////////

  /// Convenience for {@link #readVal} as Grid.
  @override
  HGrid readGrid() => readVal() as HGrid;

  /// Read a list of grids separated by blank line from stream.
  List<HGrid> readGrids() {
    var acc = <HGrid>[];
    while (_cur == HaystackToken.id) acc.add(_parseGrid());
    return acc;
  }

  /// Read a set of tags as {@code name:val} pairs separated by space.
  ///
  /// The tags may optionally be surrounded by '{' and '}'.
  HDict readDict() => _parseDict();


//////////////////////////////////////////////////////////////////////////
// Utils
//////////////////////////////////////////////////////////////////////////

  HVal _parseVal() {
    // if it's an id
    if (_cur == HaystackToken.id) {
      String id = _curVal as String;
      _consume(HaystackToken.id);

      // check for coord or xstr
      if (_cur == HaystackToken.lparen) {
        if (_peek == HaystackToken.num)
          return _parseCoord(id);
        else
          return _parseXStr(id);
      }

      // check for keyword
      if ("T" == id)   return HBool.TRUE;
      if ("F" == id)   return HBool.FALSE;
      if ("N" == id)   return null;
      if ("M" == id)   return HMarker.VAL;
      if ("NA" == id)  return HNA.VAL;
      if ("R" == id)   return HRemove.VAL;
      if ("NaN" == id) return HNum.NaN;
      if ("INF" == id) return HNum.POS_INF;

      throw _err("Unexpected identifier: $id");
    }

    // literals
    if (_cur.literal) return _parseLiteral();

    // -INF
    if (_cur == HaystackToken.minus && "INF" == _peekVal) {
      _consume(HaystackToken.minus);
      _consume(HaystackToken.id);
      return HNum.NEG_INF;
    }

    // nested collections
    if (_cur == HaystackToken.lbracket) return _parseList();
    if (_cur == HaystackToken.lbrace) return _parseDict();
    if (_cur == HaystackToken.lt2) return _parseGrid();

    throw _err("Unexpected token: " + _curToStr());
  }

  HCoord _parseCoord(String id) {
    if ("C" != id) throw _err("Expecting 'C' for coord, not $id");
    _consume(HaystackToken.lparen);
    HNum lat = _consumeNum();
    _consume(HaystackToken.comma);
    HNum lng = _consumeNum();
    _consume(HaystackToken.rparen);
    return new HCoord(lat.val, lng.val);
  }

  HVal _parseXStr(String id) {
    if (!unicode.isUppercase(id.codeUnitAt(0)))
      throw _err("Invalid XStr type: $id");
    _consume(HaystackToken.lparen);
    if (this._version < 3 && "Bin" == id) return _parseBinObsolete();
    String val = _consumeStr();
    _consume(HaystackToken.rparen);
    return HXStr.decode(id, val);
  }

  HBin _parseBinObsolete() {
    StringBuffer s = new StringBuffer();
    while (_cur != HaystackToken.rparen && _cur != HaystackToken.eof) {
      if (_curVal == null) s.write(_cur.symbol);
      else s.write(_curVal);
      _consume();
    }
    _consume(HaystackToken.rparen);
    return new HBin(s.toString());
  }

  HVal _parseLiteral() {
    Object val = this._curVal;
    if (_cur == HaystackToken.ref && _peek == HaystackToken.str) {
      val = new HRef((val as HRef).val, (_peekVal as HStr).val);
      _consume(HaystackToken.ref);
    }
    _consume();
    return val as HVal;
  }

  HList _parseList() {
    var arr = <HVal>[];

    _consume(HaystackToken.lbracket);

    while (_cur != HaystackToken.rbracket && _cur != HaystackToken.eof) {
      var val = _parseVal();
      arr.add(val);
      if (_cur != HaystackToken.comma) break;
      _consume(HaystackToken.comma);
    }
    
    _consume(HaystackToken.rbracket);

    return new HList(arr);
  }

  HDict _parseDict() {
    HDictBuilder db = new HDictBuilder();
    bool braces = _cur == HaystackToken.lbrace;
    if (braces) _consume(HaystackToken.lbrace);
    while (_cur == HaystackToken.id) {
      // tag name
      String id = _consumeTagName();
      if (!unicode.isLowercase(id.codeUnitAt(0))) throw _err("Invalid dict tag name: $id");

      // tag value
      HVal val = HMarker.VAL;
      if (_cur == HaystackToken.colon) {
        _consume(HaystackToken.colon);
        val = _parseVal();
      }
      // val returned from parseVal could be N, do not add it at all
      if ( val != null ) db.add(id, val);
    }
    if (braces) _consume(HaystackToken.rbrace);
    return db.toDict();
  }

  HGrid _parseGrid() {
    bool nested = _cur == HaystackToken.lt2;

    if (nested) {
      _consume(HaystackToken.lt2);
      if (_cur == HaystackToken.nl) _consume(HaystackToken.nl);
    }

    // ver:"3.0"
    if (_cur != HaystackToken.id || "ver" != _curVal)
      throw _err("Expecting grid 'ver' identifier, not " + _curToStr());
    _consume();
    _consume(HaystackToken.colon);
    this._version = _checkVersion(_consumeStr());

    // grid meta
    HGridBuilder gb = new HGridBuilder();

    if (_cur == HaystackToken.id)
      gb.meta.addDict(_parseDict());
    _consume(HaystackToken.nl);

    // column definitions
    int numCols = 0;

    while (_cur == HaystackToken.id) {
      ++numCols;
      String name = _consumeTagName();
      HDict colMeta = HDict.EMPTY;

      if (_cur == HaystackToken.id)
        colMeta = _parseDict();
      gb.addCol(name).addDict(colMeta);

      if (_cur != HaystackToken.comma) break;
      _consume(HaystackToken.comma);
    }

    if (numCols == 0) throw _err("No columns defined");
    _consume(HaystackToken.nl);

    // grid rows
    while (true) {
      if (_cur == HaystackToken.nl) break;
      if (_cur == HaystackToken.eof) break;
      if (nested && _cur == HaystackToken.gt2) break;

      // read cells
      List<HVal> cells = new List(numCols);

      for (int i = 0; i < numCols; ++i) {
        if (_cur == HaystackToken.comma || _cur == HaystackToken.nl || _cur == HaystackToken.eof)
          cells[i] = null;
        else
          cells[i] = _parseVal();
        if (i+1 < numCols) _consume(HaystackToken.comma);
      }

      gb.addRow(cells);

      // newline or end
      if (nested && _cur == HaystackToken.gt2) break;
      if (_cur == HaystackToken.eof) break;
      _consume(HaystackToken.nl);
    }

    if (_cur == HaystackToken.nl) _consume(HaystackToken.nl);
    if (nested) _consume(HaystackToken.gt2);
    return gb.toGrid();
  }

  int _checkVersion(String s) {
    if ("3.0" == s) return 3;
    if ("2.0" == s) return 2;
    throw _err("Unsupported version $s");
  }

//////////////////////////////////////////////////////////////////////////
// Token Reads
//////////////////////////////////////////////////////////////////////////

  String _consumeTagName() {
    _verify(HaystackToken.id);
    String id = _curVal as String;

    if (id.isEmpty || !unicode.isLowercase(id.codeUnitAt(0)))
      throw _err("Invalid dict tag name: $id");
    _consume(HaystackToken.id);
    return id;
  }

  HNum _consumeNum() {
    _verify(HaystackToken.num);
    HNum num = _curVal as HNum;
    _consume(HaystackToken.num);
    return num;
  }

  String _consumeStr() {
    _verify(HaystackToken.str);
    String val = (_curVal as HStr).val;
    _consume(HaystackToken.str);
    return val;
  }

  ParseError _err(String msg) => new ParseError(msg);

  void _verify(HaystackToken expected) {
    if (_cur != expected) throw _err("Expected $expected not " + _curToStr());
  }

  String _curToStr() => _curVal != null ? "$_cur $_curVal" : _cur.toString();

  void _consume([HaystackToken expected]) {
    if (expected != null) _verify(expected);

    _cur = _peek;
    _curVal = _peekVal;

    _peek =_tokenizer.next();
    _peekVal = _tokenizer.val;
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  HaystackTokenizer _tokenizer;

  HaystackToken _cur;
  Object _curVal;

  HaystackToken _peek;
  Object _peekVal;

  int _version = 3;
}

==> haystack-0.6.0-dev.0.1%2B2.tar.gz: lib/src/io/HZincReader.dart <==
///
/// Copyright (c) 2018, BAS Services & Graphics, LLC.
/// Licensed under the Academic Free License version 3.0
///

library haystack.hzincreader;

import 'package:unicode_helper/unicode_helper.dart' as unicode;

import '../HBin.dart';
import '../HBool.dart';
import '../HCoord.dart';
import '../HDictBuilder.dart';
import '../HGrid.dart';
import '../HGridBuilder.dart';
import '../HList.dart';
import '../HMarker.dart';
import '../HNA.dart';
import '../HNum.dart';
import '../HRef.dart';
import '../HRemove.dart';
import '../HStr.dart';
import '../HVal.dart';
import '../HXStr.dart';
import '../ParseError.dart';
import 'HGridReader.dart';
import 'HaystackToken.dart';
import 'HaystackTokenizer.dart';

/// HZincReader reads grids using the Zinc format.
///
/// See [Project Haystack](http://project-haystack.org/doc/Zinc)
class HZincReader implements HGridReader {
  HZincReader(String s) : this._tokenizer = new HaystackTokenizer(s.runes) {
    _init();
  }

  void _init() {
    _consume();
    _consume();
  }

  /// Read a value.
  HVal readVal() {
    HVal val = null;
    if (_cur == HaystackToken.id && "ver" == _curVal)
      val = _parseGrid();
    else
      val = _parseVal();
    _verify(HaystackToken.eof);
    return val;
  }

//////////////////////////////////////////////////////////////////////////
// HGridReader
//////////////////////////////////////////////////////////////////////////

  /// Convenience for {@link #readVal} as Grid.
  @override
  HGrid readGrid() => readVal() as HGrid;

  /// Read a list of grids separated by blank line from stream.
  List<HGrid> readGrids() {
    var acc = <HGrid>[];
    while (_cur == HaystackToken.id) acc.add(_parseGrid());
    return acc;
  }

  /// Read a set of tags as {@code name:val} pairs separated by space.
  ///
  /// The tags may optionally be surrounded by '{' and '}'.
  HDict readDict() => _parseDict();

//////////////////////////////////////////////////////////////////////////
// Utils
//////////////////////////////////////////////////////////////////////////

  HVal _parseVal() {
    // if it's an id
    if (_cur == HaystackToken.id) {
      String id = _curVal as String;
      _consume(HaystackToken.id);

      // check for coord or xstr
      if (_cur == HaystackToken.lparen) {
        if (_peek == HaystackToken.num)
          return _parseCoord(id);
        else
          return _parseXStr(id);
      }

      // check for keyword
      if ("T" == id) return HBool.TRUE;
      if ("F" == id) return HBool.FALSE;
      if ("N" == id) return null;
      if ("M" == id) return HMarker.VAL;
      if ("NA" == id) return HNA.VAL;
      if ("R" == id) return HRemove.VAL;
      if ("NaN" == id) return HNum.NaN;
      if ("INF" == id) return HNum.POS_INF;

      throw _err("Unexpected identifier: $id");
    }

    // literals
    if (_cur.literal) return _parseLiteral();

    // -INF
    if (_cur == HaystackToken.minus && "INF" == _peekVal) {
      _consume(HaystackToken.minus);
      _consume(HaystackToken.id);
      return HNum.NEG_INF;
    }

    // nested collections
    if (_cur == HaystackToken.lbracket) return _parseList();
    if (_cur == HaystackToken.lbrace) return _parseDict();
    if (_cur == HaystackToken.lt2) return _parseGrid();

    throw _err("Unexpected token: " + _curToStr());
  }

  HCoord _parseCoord(String id) {
    if ("C" != id) throw _err("Expecting 'C' for coord, not $id");
    _consume(HaystackToken.lparen);
    HNum lat = _consumeNum();
    _consume(HaystackToken.comma);
    HNum lng = _consumeNum();
    _consume(HaystackToken.rparen);
    return new HCoord(lat.val, lng.val);
  }

  HVal _parseXStr(String id) {
    if (!unicode.isUppercase(id.codeUnitAt(0)))
      throw _err("Invalid XStr type: $id");
    _consume(HaystackToken.lparen);
    if (this._version < 3 && "Bin" == id) return _parseBinObsolete();
    String val = _consumeStr();
    _consume(HaystackToken.rparen);
    return HXStr.decode(id, val);
  }

  HBin _parseBinObsolete() {
    StringBuffer s = new StringBuffer();
    while (_cur != HaystackToken.rparen && _cur != HaystackToken.eof) {
      if (_curVal == null)
        s.write(_cur.symbol);
      else
        s.write(_curVal);
      _consume();
    }
    _consume(HaystackToken.rparen);
    return new HBin(s.toString());
  }

  HVal _parseLiteral() {
    Object val = this._curVal;
    if (_cur == HaystackToken.ref && _peek == HaystackToken.str) {
      val = new HRef((val as HRef).val, (_peekVal as HStr).val);
      _consume(HaystackToken.ref);
    }
    _consume();
    return val as HVal;
  }

  HList _parseList() {
    var arr = <HVal>[];

    _consume(HaystackToken.lbracket);

    while (_cur != HaystackToken.rbracket && _cur != HaystackToken.eof) {
      var val = _parseVal();
      arr.add(val);
      if (_cur != HaystackToken.comma) break;
      _consume(HaystackToken.comma);
    }

    _consume(HaystackToken.rbracket);

    return new HList(arr);
  }

  HDict _parseDict() {
    HDictBuilder db = new HDictBuilder();
    bool braces = _cur == HaystackToken.lbrace;
    if (braces) _consume(HaystackToken.lbrace);
    while (_cur == HaystackToken.id) {
      // tag name
      String id = _consumeTagName();
      if (!unicode.isLowercase(id.codeUnitAt(0)))
        throw _err("Invalid dict tag name: $id");

      // tag value
      HVal val = HMarker.VAL;
      if (_cur == HaystackToken.colon) {
        _consume(HaystackToken.colon);
        val = _parseVal();
      }
      // val returned from parseVal could be N, do not add it at all
      if (val != null) db.add(id, val);
    }
    if (braces) _consume(HaystackToken.rbrace);
    return db.toDict();
  }

  HGrid _parseGrid() {
    bool nested = _cur == HaystackToken.lt2;

    if (nested) {
      _consume(HaystackToken.lt2);
      if (_cur == HaystackToken.nl) _consume(HaystackToken.nl);
    }

    // ver:"3.0"
    if (_cur != HaystackToken.id || "ver" != _curVal)
      throw _err("Expecting grid 'ver' identifier, not " + _curToStr());
    _consume();
    _consume(HaystackToken.colon);
    this._version = _checkVersion(_consumeStr());

    // grid meta
    HGridBuilder gb = new HGridBuilder();

    if (_cur == HaystackToken.id) gb.meta.addDict(_parseDict());
    _consume(HaystackToken.nl);

    // column definitions
    int numCols = 0;

    while (_cur == HaystackToken.id) {
      ++numCols;
      String name = _consumeTagName();
      HDict colMeta = HDict.EMPTY;

      if (_cur == HaystackToken.id) colMeta = _parseDict();
      gb.addCol(name).addDict(colMeta);

      if (_cur != HaystackToken.comma) break;
      _consume(HaystackToken.comma);
    }

    if (numCols == 0) throw _err("No columns defined");
    _consume(HaystackToken.nl);

    // grid rows
    while (true) {
      if (_cur == HaystackToken.nl) break;
      if (_cur == HaystackToken.eof) break;
      if (nested && _cur == HaystackToken.gt2) break;

      // read cells
      List<HVal> cells = new List(numCols);

      for (int i = 0; i < numCols; ++i) {
        if (_cur == HaystackToken.comma ||
            _cur == HaystackToken.nl ||
            _cur == HaystackToken.eof)
          cells[i] = null;
        else
          cells[i] = _parseVal();
        if (i + 1 < numCols) _consume(HaystackToken.comma);
      }

      gb.addRow(cells);

      // newline or end
      if (nested && _cur == HaystackToken.gt2) break;
      if (_cur == HaystackToken.eof) break;
      _consume(HaystackToken.nl);
    }

    if (_cur == HaystackToken.nl) _consume(HaystackToken.nl);
    if (nested) _consume(HaystackToken.gt2);
    return gb.toGrid();
  }

  int _checkVersion(String s) {
    if ("3.0" == s) return 3;
    if ("2.0" == s) return 2;
    throw _err("Unsupported version $s");
  }

//////////////////////////////////////////////////////////////////////////
// Token Reads
//////////////////////////////////////////////////////////////////////////

  String _consumeTagName() {
    _verify(HaystackToken.id);
    String id = _curVal as String;

    if (id.isEmpty || !unicode.isLowercase(id.codeUnitAt(0)))
      throw _err("Invalid dict tag name: $id");
    _consume(HaystackToken.id);
    return id;
  }

  HNum _consumeNum() {
    _verify(HaystackToken.num);
    HNum num = _curVal as HNum;
    _consume(HaystackToken.num);
    return num;
  }

  String _consumeStr() {
    _verify(HaystackToken.str);
    String val = (_curVal as HStr).val;
    _consume(HaystackToken.str);
    return val;
  }

  ParseError _err(String msg) => new ParseError(msg);

  void _verify(HaystackToken expected) {
    if (_cur != expected) throw _err("Expected $expected not " + _curToStr());
  }

  String _curToStr() => _curVal != null ? "$_cur $_curVal" : _cur.toString();

  void _consume([HaystackToken expected]) {
    if (expected != null) _verify(expected);

    _cur = _peek;
    _curVal = _peekVal;

    _peek = _tokenizer.next();
    _peekVal = _tokenizer.val;
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  HaystackTokenizer _tokenizer;

  HaystackToken _cur;
  Object _curVal;

  HaystackToken _peek;
  Object _peekVal;

  int _version = 3;
}
